from dagster._core.libraries import DagsterLibraryRegistry

from dagster_sdf.asset_decorator import sdf_assets as sdf_assets
from dagster_sdf.asset_utils import (
    build_schedule_from_sdf_selection as build_schedule_from_sdf_selection,
    dagster_name_fn as dagster_name_fn,
    default_asset_key_fn as default_asset_key_fn,
)
from dagster_sdf.resource import SdfCliResource as SdfCliResource
from dagster_sdf.sdf_information_schema import SdfInformationSchema as SdfInformationSchema
from dagster_sdf.sdf_workspace import SdfWorkspace as SdfWorkspace
from dagster_sdf.version import __version__ as __version__

DagsterLibraryRegistry.register("dagster-sdf", __version__)



import logging
import os
from pathlib import Path
from typing import Optional, Sequence, Union

from dagster._annotations import experimental
from dagster._model import DagsterModel
from dagster._utils import run_with_concurrent_update_guard
from dagster._utils.warnings import suppress_dagster_warnings

from dagster_sdf.constants import DEFAULT_SDF_WORKSPACE_ENVIRONMENT, SDF_EXECUTABLE, SDF_TARGET_DIR
from dagster_sdf.errors import (
    DagsterSdfInformationSchemaNotFoundError,
    DagsterSdfWorkspaceNotFoundError,
    DagsterSdfWorkspaceYmlFileNotFoundError,
)
from dagster_sdf.sdf_information_schema import SdfInformationSchema

logger = logging.getLogger("dagster-sdf.artifacts")


def using_dagster_dev() -> bool:
    return bool(os.getenv("DAGSTER_IS_DEV_CLI"))


@experimental
class SdfWorkspacePreparer:
    def on_load(self, workspace: "SdfWorkspace") -> None:
        """Invoked when SdfWorkspace is instantiated with this preparer."""

    def prepare(self, workspace: "SdfWorkspace") -> None:
        """Called explictly to compile the workspace."""

    def using_dagster_dev(self) -> bool:
        return using_dagster_dev()

    def compile_on_load_opt_in(self) -> bool:
        return bool(os.getenv("DAGSTER_SDF_COMPILE_ON_LOAD"))


@experimental
class DagsterSdfWorkspacePreparer(SdfWorkspacePreparer):
    def __init__(
        self,
        generate_cli_args: Optional[Sequence[str]] = None,
    ):
        """The default SdfWorkspacePreparer, this handler provides an experience of:
            * During development, rerun compile to pick up any changes.
            * When deploying, expect the outcome of compile to reduce start-up time.

        Args:
            generate_cli_args (Sequence[str]):
                The arguments to pass to the sdf cli to prepare the workspace.
                Default: ["compile", "--save==table-deps,info-schema"]
        """
        self._generate_cli_args = generate_cli_args or [
            "compile",
            "--save",
            "table-deps,info-schema",
        ]

    def on_load(self, workspace: "SdfWorkspace"):
        if self.using_dagster_dev() or self.compile_on_load_opt_in():
            self.prepare(workspace)
            information_schema = SdfInformationSchema(
                workspace_dir=workspace.workspace_dir,
                target_dir=workspace.target_dir,
                environment=workspace.environment,
            )
            if not information_schema.is_parsed():
                raise DagsterSdfInformationSchemaNotFoundError(
                    f"Sdf Information Schema was not generated correctly at expected path {information_schema.information_schema_dir} "
                    f"after running '{self.prepare.__qualname__}'. Ensure the implementation respects "
                    "all Sdf Workspace properties."
                )

    def prepare(self, workspace: "SdfWorkspace") -> None:
        output_dir = workspace.target_dir.joinpath(SDF_TARGET_DIR, workspace.environment)
        run_with_concurrent_update_guard(
            output_dir,
            self._prepare_workspace,
            workspace=workspace,
        )

    def _prepare_workspace(self, workspace: "SdfWorkspace") -> None:
        from dagster_sdf.resource import SdfCliResource

        (
            SdfCliResource(workspace_dir=workspace)
            .cli(
                self._generate_cli_args,
                target_dir=workspace.target_dir,
                environment=workspace.environment,
                raise_on_error=False,
            )
            .wait()
        )


@suppress_dagster_warnings
@experimental
class SdfWorkspace(DagsterModel):
    """The SdfWorkspace is a representation of an sdf workspace that can be compiled and executed.

    Args:
    workspace_dir (Union[str, Path]):
        The directory of the sdf workspace.
    target_dir (Union[str, Path]):
        The directory, relative to the workspace directory, to target and output artifacts.
        Default: "sdf_dagster_target"
    environment (str):
        The environment to use for compilation and execution. Default: "dbg"
    workspace_preparer (Optional[SdfWorkspacePreparer]):
        A object for ensuring that the sdf workspace is in the right state at the right times.
        Default: DagsterSdfWorkspacePreparer
    sdf_executable (str):
        The path/name of the sdf executable to use. Default: "sdf"
    """

    workspace_dir: Path
    target_dir: Path
    environment: str
    workspace_preparer: SdfWorkspacePreparer
    sdf_executable: str

    def __init__(
        self,
        workspace_dir: Union[str, Path],
        *,
        target_dir: Union[str, Path] = Path(
            ""
        ),  # TODO: This should default to `sdftarget` but it will cause a double-up of `sdftarget` in the path
        environment: str = DEFAULT_SDF_WORKSPACE_ENVIRONMENT,
        workspace_preparer: SdfWorkspacePreparer = DagsterSdfWorkspacePreparer(),
        sdf_executable: str = SDF_EXECUTABLE,
    ):
        workspace_dir = Path(workspace_dir)
        if not workspace_dir.exists():
            raise DagsterSdfWorkspaceNotFoundError(f"workspace_dir {workspace_dir} does not exist.")

        target_dir = workspace_dir.joinpath(target_dir)

        sdf_workspace_yml_path = workspace_dir.joinpath("workspace.sdf.yml")
        if not sdf_workspace_yml_path.exists():
            raise DagsterSdfWorkspaceYmlFileNotFoundError(
                f"Did not find workspace.sdf.yml at expected path {sdf_workspace_yml_path}. "
                f"Ensure the specified workspace directory respects all sdf workspace requirements."
            )

        super().__init__(
            workspace_dir=workspace_dir,
            target_dir=target_dir,
            environment=environment,
            workspace_preparer=workspace_preparer,
            sdf_executable=sdf_executable,
        )
        if workspace_preparer:
            workspace_preparer.on_load(self)



import textwrap
from pathlib import Path
from typing import AbstractSet, Any, Dict, Mapping, Optional, Sequence

from dagster import (
    AssetCheckKey,
    AssetKey,
    AssetsDefinition,
    AssetSelection,
    DefaultScheduleStatus,
    RunConfig,
    ScheduleDefinition,
    define_asset_job,
)

from dagster_sdf.constants import SDF_TARGET_DIR


def dagster_name_fn(catalog: str, schema: str, table: str) -> str:
    fmt_catalog = catalog.replace(".", "_").replace("-", "_").replace("*", "_star")
    fmt_schema = schema.replace(".", "_").replace("-", "_").replace("*", "_star")
    fmt_table = table.replace(".", "_").replace("-", "_").replace("*", "_star")
    return f"{fmt_catalog}__{fmt_schema}__{fmt_table}"


def default_asset_key_fn(catalog: str, schema: str, table: str) -> AssetKey:
    """Get the asset key for an sdf asset. An Sdf asset's key is its fully qualified name."""
    return AssetKey([catalog, schema, table])


def default_asset_check_key_fn(catalog: str, schema: str, table: str) -> AssetCheckKey:
    """Get the asset check key for an sdf asset. An Sdf asset's check key is its fully qualified name."""
    # If table_name starts with "test_", this is a column or table test (to be updated with something more formal)
    if table.lower().startswith("test_"):
        asset_key = AssetKey([catalog, schema, table[5:]])
    else:
        asset_key = AssetKey([catalog, schema, table])
    return AssetCheckKey(
        name="test",
        asset_key=asset_key,
    )


def build_schedule_from_sdf_selection(
    _: Sequence[AssetsDefinition],
    job_name: str,
    cron_schedule: str,
    schedule_name: Optional[str] = None,
    tags: Optional[Mapping[str, str]] = None,
    config: Optional[RunConfig] = None,
    execution_timezone: Optional[str] = None,
    default_status: DefaultScheduleStatus = DefaultScheduleStatus.STOPPED,
) -> ScheduleDefinition:
    """Build a schedule to materialize a specified set of sdf models (Currently Limited to All).

    Args:
        job_name (str): The name of the job to materialize the sdf models.
        cron_schedule (str): The cron schedule to define the schedule.
        schedule_name (Optional[str]): The name of the sdf schedule to create.
        tags (Optional[Mapping[str, str]]): A dictionary of tags (string key-value pairs) to attach
            to the scheduled runs.
        config (Optional[RunConfig]): The config that parameterizes the execution of this schedule.
        execution_timezone (Optional[str]): Timezone in which the schedule should run.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>` - e.g. "America/Los_Angeles".

    Returns:
        ScheduleDefinition: A definition to materialize the selected sdf models on a cron schedule.

    Examples:
        .. code-block:: python

            from dagster_sdf import sdf_assets, build_schedule_from_sdf_selection

            @sdf_assets(workspace=...)
            def all_sdf_assets():
                ...

            daily_sdf_assets_schedule = build_schedule_from_sdf_selection(
                [all_sdf_assets],
                job_name="all_sdf_assets",
                cron_schedule="0 0 * * *",
            )
    """
    return ScheduleDefinition(
        name=schedule_name,
        cron_schedule=cron_schedule,
        job=define_asset_job(
            name=job_name,
            selection=AssetSelection.all(),
            config=config,
            tags=tags,
        ),
        execution_timezone=execution_timezone,
        default_status=default_status,
    )


def get_asset_key_for_table_id(sdf_assets: Sequence[AssetsDefinition], table_id: str) -> AssetKey:
    """Returns the corresponding Dagster asset key for an sdf table.

    Args:
        table_id (str): The fqn of the the sdf table.

    Raises:
        DagsterInvalidInvocationError: If the source has more than one table.

    Returns:
        AssetKey: The corresponding Dagster asset key.

    Examples:
        .. code-block:: python

            from dagster import asset
            from dagster_sdf import sdf_assets, get_asset_key_for_table_id

            @sdf_assets(workspace=...)
            def all_sdf_assets():
                ...

            @asset(key=get_asset_key_for_table_id([all_sdf_assets], "my_source"))
            def upstream_python_asset():
                ...
    """
    # Strip table_id of any quotes
    table_id = table_id.replace('"', "").replace("`", "")
    asset_keys_by_output_name = {}
    for sdf_asset in sdf_assets:
        for name, key in sdf_asset.keys_by_output_name.items():
            key_table_id = ".".join(key.path)
            if key_table_id == table_id:
                asset_keys_by_output_name[name] = key

    if len(asset_keys_by_output_name) > 1:
        raise KeyError(
            f"Table {table_id} has more than one table associated with two or more unique dagster asset names:"
            f" {asset_keys_by_output_name.keys()}."
        )
    if not asset_keys_by_output_name:
        raise KeyError(f"Table {table_id} not found in the provided sdf assets.")
    return next(iter(asset_keys_by_output_name.values()))


def get_output_dir(target_dir: Path, environment: str) -> Path:
    return target_dir.joinpath(SDF_TARGET_DIR, environment)


def get_info_schema_dir(target_dir: Path, environment: str) -> Path:
    return get_output_dir(target_dir, environment).joinpath(
        "data", "system", "information_schema::sdf"
    )


def get_materialized_sql_dir(target_dir: Path, environment: str) -> Path:
    return get_output_dir(target_dir, environment).joinpath("materialized")


def get_table_path_from_parts(catalog_name: str, schema_name: str, table_name: str) -> Path:
    return Path(catalog_name.lower(), schema_name.lower(), table_name.lower())


def _read_sql_file(path_to_file: Path) -> str:
    with open(path_to_file, "r") as file:
        return textwrap.indent(file.read().strip(), "    ")


def default_description_fn(
    table_row: Dict[str, Any],
    workspace_dir: Optional[Path] = None,
    output_dir: Optional[Path] = None,
    enable_raw_sql_description: bool = True,
    enable_materialized_sql_description: bool = True,
) -> str:
    description_sections = [
        table_row["description"] or f"sdf {table_row['materialization']} {table_row['table_id']}",
    ]
    if enable_raw_sql_description:
        if workspace_dir is None:
            raise ValueError("workspace_dir must be provided to enable raw SQL description.")
        path_to_file = None
        for source_location in table_row["source_locations"]:
            if source_location.endswith(".sql"):
                path_to_file = workspace_dir.joinpath(source_location)
                break
        if path_to_file is not None and path_to_file.exists():
            description_sections.append(f"#### Raw SQL:\n```\n{_read_sql_file(path_to_file)}\n```")
    if enable_materialized_sql_description:
        if output_dir is None:
            raise ValueError("output_dir must be provided to enable materialized SQL description.")
        path_to_file = (
            output_dir.joinpath("materialized")
            .joinpath(
                get_table_path_from_parts(
                    table_row["catalog_name"], table_row["schema_name"], table_row["table_name"]
                )
            )
            .with_suffix(".sql")
        )
        if path_to_file.exists():
            description_sections.append(
                f"#### Materialized SQL:\n```\n{_read_sql_file(path_to_file)}\n```"
            )
    return "\n\n".join(filter(None, description_sections))


def get_test_prefix(table_dialect: str) -> str:
    if table_dialect == "snowflake":
        return "TEST_"
    else:
        return "test_"


def exists_in_selected(
    catalog_name: str,
    schema_name: str,
    table_name: str,
    purpose: str,
    dialect: str,
    selected_output_names: AbstractSet[str],
    asset_checks_enabled: bool,
) -> bool:
    asset_output_name = dagster_name_fn(catalog_name, schema_name, table_name)
    # If asset checks are enabled, ensure tests are only yielded for selected assets
    if asset_checks_enabled:
        # Strip test_name_prefix from start of table_id if it exists
        if purpose == "test":
            test_name_prefix = get_test_prefix(dialect)
            asset_output_name = (
                dagster_name_fn(catalog_name, schema_name, table_name[len(test_name_prefix) :])
                if table_name.startswith(test_name_prefix)
                else asset_output_name
            )

    if asset_output_name in selected_output_names:
        return True
    else:
        return False



import os
import re
import shutil
import subprocess
import uuid
from contextlib import suppress
from pathlib import Path
from typing import Any, List, Optional, Sequence, Union

from dagster import (
    AssetExecutionContext,
    AssetsDefinition,
    ConfigurableResource,
    DagsterInvariantViolationError,
    OpExecutionContext,
    get_dagster_logger,
)
from dagster._annotations import experimental, public
from dagster._core.errors import DagsterInvalidPropertyError
from dagster._utils.warnings import suppress_dagster_warnings
from pydantic import Field, validator

from dagster_sdf.asset_utils import dagster_name_fn, get_test_prefix
from dagster_sdf.constants import (
    DAGSTER_SDF_CATALOG_NAME,
    DAGSTER_SDF_DIALECT,
    DAGSTER_SDF_SCHEMA_NAME,
    DAGSTER_SDF_TABLE_ID,
    DAGSTER_SDF_TABLE_NAME,
    DEFAULT_SDF_WORKSPACE_ENVIRONMENT,
    SDF_DAGSTER_OUTPUT_DIR,
    SDF_EXECUTABLE,
    SDF_WORKSPACE_YML,
)
from dagster_sdf.dagster_sdf_translator import DagsterSdfTranslator, validate_opt_translator
from dagster_sdf.sdf_cli_invocation import SdfCliInvocation
from dagster_sdf.sdf_version import SDF_VERSION_LOWER_BOUND, SDF_VERSION_UPPER_BOUND
from dagster_sdf.sdf_workspace import SdfWorkspace

logging = get_dagster_logger()


@suppress_dagster_warnings
@experimental
class SdfCliResource(ConfigurableResource):
    """A resource used to execute sdf CLI commands.

    Attributes:
        workspace_dir (str): The path to the sdf workspace directory. This directory should contain a
            `workspace.sdf.yml`.
        global_config_flags (List[str]): A list of global flags configuration to pass to the sdf CLI
            invocation.
        sdf_executable (str): The path to the sdf executable. By default, this is `sdf`.
    """

    workspace_dir: str = Field(
        description=(
            "The path to the sdf workspace directory. This directory should contain a"
            " `workspace.sdf.yml`."
        ),
    )
    global_config_flags: List[str] = Field(
        default=[],
        description="A list of global flags configuration to pass to the sdf CLI invocation.",
    )
    sdf_executable: str = Field(
        default=SDF_EXECUTABLE,
        description="The path to the sdf executable.",
    )

    def __init__(
        self,
        workspace_dir: Union[str, Path, SdfWorkspace],
        global_config_flags: Optional[List[str]] = None,
        sdf_executable: str = SDF_EXECUTABLE,
        **kwargs,  # allow custom subclasses to add fields
    ):
        if isinstance(workspace_dir, SdfWorkspace):
            workspace_dir = workspace_dir.workspace_dir
        # static typing doesn't understand whats going on here, thinks these fields dont exist
        super().__init__(
            workspace_dir=workspace_dir,  # type: ignore
            global_config_flags=global_config_flags or [],  # type: ignore
            sdf_executable=sdf_executable,  # type: ignore
            **kwargs,
        )

    @validator("workspace_dir", pre=True)
    def convert_path_to_str(cls, v: Any) -> Any:
        """Validate that the path is converted to a string."""
        if isinstance(v, Path):
            resolved_path = cls._validate_absolute_path_exists(v)

            absolute_path = Path(v).absolute()
            try:
                resolved_path = absolute_path.resolve(strict=True)
            except FileNotFoundError:
                raise ValueError(f"The absolute path of '{v}' ('{absolute_path}') does not exist")
            return os.fspath(resolved_path)

        return v

    @validator("workspace_dir")
    def validate_workspace_dir(cls, project_dir: str) -> str:
        resolved_workspace_dir = cls._validate_absolute_path_exists(project_dir)

        cls._validate_path_contains_file(
            path=resolved_workspace_dir,
            file_name=SDF_WORKSPACE_YML,
            error_message=(
                f"{resolved_workspace_dir} does not contain an {SDF_WORKSPACE_YML} file. Please"
                " specify a valid path to an sdf workspace."
            ),
        )

        return os.fspath(resolved_workspace_dir)

    @validator("sdf_executable")
    def validate_sdf_executable(cls, sdf_executable: str) -> str:
        resolved_sdf_executable = shutil.which(sdf_executable)
        if not resolved_sdf_executable:
            raise ValueError(
                f"The sdf executable '{sdf_executable}' does not exist. Please specify a valid"
                " path to an sdf executable."
            )

        return sdf_executable

    @public
    def cli(
        self,
        args: Sequence[str],
        *,
        target_dir: Optional[Path] = None,
        environment: str = DEFAULT_SDF_WORKSPACE_ENVIRONMENT,
        raise_on_error: bool = True,
        dagster_sdf_translator: Optional[DagsterSdfTranslator] = None,
        context: Optional[Union[OpExecutionContext, AssetExecutionContext]] = None,
    ) -> SdfCliInvocation:
        """Create a subprocess to execute an sdf CLI command.

        Args:
            args (Sequence[str]): The sdf CLI command to execute.
            target_dir (Optional[Path]): The path to the target directory.
            environment (str): The environment to use. Defaults to "dbg".
            raise_on_error (bool): Whether to raise an exception if the sdf CLI command fails.
            context (Optional[Union[OpExecutionContext, AssetExecutionContext]]): The execution context from within `@sdf_assets`.
                If an AssetExecutionContext is passed, its underlying OpExecutionContext will be used.

        Returns:
            SdfCliInvocation: A invocation instance that can be used to retrieve the output of the
                sdf CLI command.
        """
        self._validate_sdf_version()
        dagster_sdf_translator = validate_opt_translator(dagster_sdf_translator)
        assets_def: Optional[AssetsDefinition] = None
        with suppress(DagsterInvalidPropertyError):
            assets_def = context.assets_def if context else None

        context = (
            context.op_execution_context if isinstance(context, AssetExecutionContext) else context
        )

        dagster_sdf_translator = dagster_sdf_translator or DagsterSdfTranslator()
        run_args = []
        # If the context and assets_def are not None, then we need to pass in the selected targets
        # to the sdf CLI invocation.
        if context and assets_def is not None:
            run_args.append(
                "--targets-only"
            )  # This flag is used to only run the selected targets (and not upstreams)
            selected_output_names = context.selected_output_names
            for asset_key, asset_metadata in assets_def.metadata_by_key.items():
                asset_output_name = dagster_name_fn(
                    asset_key.path[0], asset_key.path[1], asset_key.path[2]
                )
                # Check if this output is expected, if so, add the table_id and optionally the expected test name to the run_args
                if selected_output_names and asset_output_name in selected_output_names:
                    table_id: Optional[str] = asset_metadata.get(DAGSTER_SDF_TABLE_ID)
                    catalog: Optional[str] = asset_metadata.get(DAGSTER_SDF_CATALOG_NAME)
                    schema: Optional[str] = asset_metadata.get(DAGSTER_SDF_SCHEMA_NAME)
                    table_name: Optional[str] = asset_metadata.get(DAGSTER_SDF_TABLE_NAME)
                    table_dialect: Optional[str] = asset_metadata.get(DAGSTER_SDF_DIALECT)
                    # If any of the metadata is missing, raise an error
                    if (
                        table_id is None
                        or catalog is None
                        or schema is None
                        or table_name is None
                        or table_dialect is None
                    ):
                        raise DagsterInvariantViolationError(
                            f"Expected to find sdf table metadata on asset {asset_key.to_user_string()},"
                            " but did not. Did you pass in assets that weren't generated by @sdf_assets?"
                            " Please ensure that the asset metadata contains the following keys: "
                            f"{DAGSTER_SDF_TABLE_ID}, {DAGSTER_SDF_CATALOG_NAME}, {DAGSTER_SDF_SCHEMA_NAME}, {DAGSTER_SDF_TABLE_NAME}, {DAGSTER_SDF_DIALECT}"
                        )
                    run_args.append(table_id)
                    test_name_prefix = get_test_prefix(table_dialect)
                    # If the command is test, then we need to add the test name to the run_args (temporary: until sdf-cli applies --targets-only for test)
                    if args[0] == "test":
                        run_args.append(f"{catalog}.{schema}.{test_name_prefix}{table_name}")

        # Pass the current environment variables to the sdf CLI invocation.
        env = os.environ.copy()

        environment_args = ["--environment", environment]
        target_path = target_dir or self._get_unique_target_path(context=context)
        target_args = ["--target-dir", str(target_path)]
        log_level_args = ["--log-level", "info"]

        # Ensure that the target_dir exists
        target_path.mkdir(parents=True, exist_ok=True)

        args = [
            self.sdf_executable,
            *self.global_config_flags,
            *log_level_args,
            *args,
            *environment_args,
            *target_args,
            *run_args,
        ]

        return SdfCliInvocation.run(
            args=args,
            env=env,
            workspace_dir=Path(self.workspace_dir),
            target_dir=target_path,
            environment=environment,
            dagster_sdf_translator=dagster_sdf_translator,
            raise_on_error=raise_on_error,
            context=context,
        )

    @classmethod
    def _validate_absolute_path_exists(cls, path: Union[str, Path]) -> Path:
        absolute_path = Path(path).absolute()
        try:
            resolved_path = absolute_path.resolve(strict=True)
        except FileNotFoundError:
            raise ValueError(f"The absolute path of '{path}' ('{absolute_path}') does not exist")

        return resolved_path

    @classmethod
    def _validate_path_contains_file(cls, path: Path, file_name: str, error_message: str):
        if not path.joinpath(file_name).exists():
            raise ValueError(error_message)

    def _get_unique_target_path(self, *, context: Optional[OpExecutionContext]) -> Path:
        """Get a unique target path for the sdf CLI invocation.

        Args:
            context (Optional[OpExecutionContext]): The execution context.

        Returns:
            str: A unique target path for the sdf CLI invocation.
        """
        unique_id = str(uuid.uuid4())[:7]
        path = unique_id
        if context:
            path = f"{context.op.name}-{context.run.run_id[:7]}-{unique_id}"

        current_output_path = Path(self.workspace_dir).joinpath(SDF_DAGSTER_OUTPUT_DIR)

        return current_output_path.joinpath(path)

    def _validate_sdf_version(self) -> None:
        """Validate that the sdf version is compatible with the current version of the sdf CLI."""
        try:
            result = subprocess.run(
                ["sdf", "--version"], capture_output=True, text=True, check=True
            )
            output = result.stdout.strip()
            match = re.search(r"sdf (\d+\.\d+\.\d+)", output)
            if match:
                version = match.group(1)
                if version < SDF_VERSION_LOWER_BOUND or version >= SDF_VERSION_UPPER_BOUND:
                    logging.warn(
                        f"The sdf version '{version}' is not within the supported range of"
                        f" '{SDF_VERSION_LOWER_BOUND}' to '{SDF_VERSION_UPPER_BOUND}'. Check your"
                        " environment to ensure that the correct version of sdf is being used."
                    )
            else:
                logging.warn(
                    "Failed to extract the sdf version from the output. Check your environment to"
                    " ensure that the correct version of sdf is being used."
                )
        except subprocess.CalledProcessError as e:
            logging.warn(f"Failed to get the sdf version: {e}")
            exit(1)



import os
import shutil
from pathlib import Path
from typing import Any

import typer
import yaml
from dagster._cli.project import check_if_pypi_package_conflict_exists
from jinja2 import Environment, FileSystemLoader
from rich.console import Console
from rich.syntax import Syntax
from typing_extensions import Annotated

from dagster_sdf.include import STARTER_PROJECT_PATH
from dagster_sdf.sdf_version import SDF_VERSION_UPPER_BOUND
from dagster_sdf.version import __version__ as dagster_sdf_version

app = typer.Typer(
    no_args_is_help=True,
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
    help="CLI tools for working with Dagster and sdf.",
    add_completion=False,
)
workspace_app = typer.Typer(
    name="workspace",
    no_args_is_help=True,
    help="Commands for using an sdf workspace in Dagster.",
    add_completion=False,
)
app.add_typer(workspace_app)

console = Console()

SDF_WORKSPACE_YML_NAME = "workspace.sdf.yml"


def validate_dagster_workspace_name(workspace_name: str) -> str:
    if not workspace_name.isidentifier():
        raise typer.BadParameter(
            f"The workspace name `{workspace_name}` is not a valid sql identifier containing only"
            " letters, digits, or underscores. Please specify a valid project name."
        )

    return workspace_name


def validate_sdf_workspace_dir(sdf_workspace_dir: Path) -> Path:
    sdf_workspace_yaml_path = sdf_workspace_dir.joinpath(SDF_WORKSPACE_YML_NAME)

    if not sdf_workspace_yaml_path.exists():
        raise typer.BadParameter(
            f"{sdf_workspace_dir} does not contain a {SDF_WORKSPACE_YML_NAME} file. Please specify a"
            " valid path to an sdf workspace."
        )

    return sdf_workspace_dir


def find_workspace_block(documents: list) -> Any:
    for doc in documents:
        if "workspace" in doc:
            return doc["workspace"]
    return None


def copy_scaffold(
    workspace_name: str,
    dagster_project_dir: Path,
    sdf_workspace_dir: Path,
) -> None:
    sdf_workspace_yaml_path = sdf_workspace_dir.joinpath(SDF_WORKSPACE_YML_NAME)
    with sdf_workspace_yaml_path.open("rb") as file:
        sdf_workspace_yaml = list(yaml.safe_load_all(file))
    sdf_workspace_block = find_workspace_block(sdf_workspace_yaml)
    sdf_workspace_name: str = sdf_workspace_block["name"]

    shutil.copytree(
        src=STARTER_PROJECT_PATH,
        dst=dagster_project_dir,
        ignore=shutil.ignore_patterns("__pycache__"),
    )
    dagster_project_dir.joinpath("__init__.py").unlink()

    sdf_workspace_dir_relative_path = Path(
        os.path.relpath(
            sdf_workspace_dir,
            start=dagster_project_dir.joinpath(workspace_name, "definitions.py"),
        )
    )
    sdf_workspace_dir_relative_path_parts = [
        f'"{part}"' for part in sdf_workspace_dir_relative_path.parts
    ]
    env = Environment(loader=FileSystemLoader(dagster_project_dir))

    for path in dagster_project_dir.glob("**/*"):
        if path.suffix == ".jinja":
            relative_path = path.relative_to(Path.cwd())
            destination_path = os.fspath(relative_path.parent.joinpath(relative_path.stem))
            template_path = path.relative_to(dagster_project_dir).as_posix()

            env.get_template(template_path).stream(
                sdf_workspace_dir_relative_path_parts=sdf_workspace_dir_relative_path_parts,
                sdf_workspace_name=f"{sdf_workspace_name}_project",
                sdf_assets_name=f"{sdf_workspace_name}_sdf_assets",
                sdf_version_upper_bound=SDF_VERSION_UPPER_BOUND,
                workspace_name=workspace_name,
            ).dump(destination_path)

            path.unlink()

    dagster_project_dir.joinpath("scaffold").rename(dagster_project_dir.joinpath(workspace_name))


def _check_and_error_on_package_conflicts(workspace_name: str) -> None:
    package_check_result = check_if_pypi_package_conflict_exists(workspace_name)
    if package_check_result.request_error_msg:
        console.print(
            f"An error occurred while checking if workspace name '{workspace_name}' conflicts with"
            f" an existing PyPI package: {package_check_result.request_error_msg}."
            " \n\nConflicting package names will cause import errors in your project if the"
            " existing PyPI package is included as a dependency in your scaffolded project. If"
            " desired, this check can be skipped by adding the `--ignore-package-conflict`"
            " flag."
        )
        raise typer.Exit(1)

    if package_check_result.conflict_exists:
        raise typer.BadParameter(
            f"The workspace name '{workspace_name}' conflicts with an existing PyPI package."
            " Conflicting package names will cause import errors in your project if the"
            " existing PyPI package is included as a dependency in your scaffolded project."
            " Please choose another name, or add the `--ignore-package-conflict` flag to"
            " bypass this check."
        )


@workspace_app.command(name="scaffold")
def project_scaffold_command(
    project_name: Annotated[
        str,
        typer.Option(
            default=...,
            callback=validate_dagster_workspace_name,
            show_default=False,
            prompt="Enter a name for your Dagster project (letters, digits, underscores)",
            help="The name of the Dagster project to initialize for your sdf workspace.",
        ),
    ],
    sdf_workspace_dir: Annotated[
        Path,
        typer.Option(
            default=...,
            callback=validate_sdf_workspace_dir,
            show_default=False,
            help=(
                "The path of your sdf workspace directory. This path must contain a workspace.sdf.yml"
                " file. By default, this command will assume that the current working directory"
                " contains an sdf workspace, but you can set a different directory by setting this"
                " option."
            ),
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=True,
            resolve_path=True,
        ),
    ] = Path.cwd(),
    ignore_package_conflict: Annotated[
        bool,
        typer.Option(
            default=...,
            help="Controls whether the workspace name can conflict with an existing PyPI package.",
            is_flag=True,
            hidden=True,
        ),
    ] = False,
) -> None:
    """This command will initialize a new Dagster project and create directories and files that
    load assets from an existing sdf workspace.
    """
    if not ignore_package_conflict:
        _check_and_error_on_package_conflicts(project_name)

    console.print(
        f"Running with dagster-sdf version: [bold green]{dagster_sdf_version}[/bold green]."
    )
    console.print(
        f"Initializing Dagster project [bold green]{project_name}[/bold green] in the current"
        f" working directory for sdf workspace directory [bold green]{sdf_workspace_dir}[/bold green].",
    )

    dagster_project_dir = Path.cwd().joinpath(project_name)

    copy_scaffold(
        workspace_name=project_name,
        dagster_project_dir=dagster_project_dir,
        sdf_workspace_dir=sdf_workspace_dir,
    )

    dagster_dev_command = "DAGSTER_SDF_COMPILE_ON_LOAD=1 dagster dev"

    console.print(
        "Your Dagster project has been initialized. To view your sdf workspace in Dagster, run"
        " the following commands:",
        Syntax(
            code="\n".join(
                [
                    f"cd '{dagster_project_dir}'",
                    dagster_dev_command,
                ]
            ),
            lexer="bash",
            padding=1,
        ),
    )


project_app_typer_click_object = typer.main.get_command(workspace_app)



from typing import Any, Callable, Mapping, Optional, Set

from dagster import (
    AssetsDefinition,
    BackfillPolicy,
    PartitionsDefinition,
    RetryPolicy,
    TimeWindowPartitionsDefinition,
    multi_asset,
)
from dagster._annotations import experimental
from dagster._utils.warnings import suppress_dagster_warnings

from dagster_sdf.dagster_sdf_translator import DagsterSdfTranslator, validate_translator
from dagster_sdf.sdf_information_schema import SdfInformationSchema
from dagster_sdf.sdf_workspace import SdfWorkspace


@suppress_dagster_warnings
@experimental
def sdf_assets(
    *,
    workspace: SdfWorkspace,
    name: Optional[str] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    dagster_sdf_translator: Optional[DagsterSdfTranslator] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    op_tags: Optional[Mapping[str, Any]] = None,
    required_resource_keys: Optional[Set[str]] = None,
    retry_policy: Optional[RetryPolicy] = None,
) -> Callable[[Callable[..., Any]], AssetsDefinition]:
    dagster_sdf_translator = validate_translator(dagster_sdf_translator or DagsterSdfTranslator())
    information_schema = SdfInformationSchema(
        workspace_dir=workspace.workspace_dir,
        target_dir=workspace.target_dir,
        environment=workspace.environment,
    )
    (specs, check_specs) = information_schema.build_sdf_multi_asset_args(
        dagster_sdf_translator=dagster_sdf_translator,
    )

    if (
        partitions_def
        and isinstance(partitions_def, TimeWindowPartitionsDefinition)
        and not backfill_policy
    ):
        backfill_policy = BackfillPolicy.single_run()
    return multi_asset(
        specs=specs,
        name=name,
        required_resource_keys=required_resource_keys,
        partitions_def=partitions_def,
        can_subset=True,
        op_tags=op_tags,
        check_specs=check_specs,
        backfill_policy=backfill_policy,
        retry_policy=retry_policy,
    )



from collections import abc
from typing import TYPE_CHECKING, Generic, Iterator, Union

from dagster import AssetCheckResult, AssetMaterialization, AssetObservation, Output
from typing_extensions import TypeVar

if TYPE_CHECKING:
    from dagster_sdf.sdf_cli_invocation import SdfCliInvocation

SdfDagsterEventType = Union[Output, AssetMaterialization, AssetObservation, AssetCheckResult]

# We define SdfEventIterator as a generic type for the sake of type hinting.
# This is so that users who inspect the type of the return value of `SdfCliInvocation.stream()`
# will be able to see the inner type of the iterator, rather than just `SdfEventIterator`.
T = TypeVar("T", bound=SdfDagsterEventType)


class SdfEventIterator(Generic[T], abc.Iterator):
    """A wrapper around an iterator of sdf events which contains additional methods for
    post-processing the events.
    """

    def __init__(
        self,
        exec_events: Iterator[T],
        sdf_cli_invocation: "SdfCliInvocation",
    ) -> None:
        self._inner_iterator = exec_events
        self._sdf_cli_invocation = sdf_cli_invocation

    def __next__(self) -> T:
        return next(self._inner_iterator)

    def __iter__(self) -> "SdfEventIterator[T]":
        return self



from dataclasses import dataclass
from typing import AbstractSet, Any, Dict, Iterator, Optional

from dagster import (
    AssetCheckResult,
    AssetCheckSeverity,
    AssetMaterialization,
    OpExecutionContext,
    Output,
)
from dagster._annotations import public

from dagster_sdf.asset_utils import dagster_name_fn, exists_in_selected
from dagster_sdf.constants import (
    DAGSTER_SDF_CATALOG_NAME,
    DAGSTER_SDF_DIALECT,
    DAGSTER_SDF_PURPOSE,
    DAGSTER_SDF_SCHEMA_NAME,
    DAGSTER_SDF_TABLE_ID,
    DAGSTER_SDF_TABLE_NAME,
)
from dagster_sdf.dagster_sdf_translator import DagsterSdfTranslator
from dagster_sdf.sdf_event_iterator import SdfDagsterEventType


@dataclass
class SdfCliEventMessage:
    """The representation of an sdf CLI event.

    Args:
        raw_event (Dict[str, Any]): The raw event dictionary.
    """

    raw_event: Dict[str, Any]

    @property
    def is_result_event(self) -> bool:
        return (
            bool(self.raw_event.get("ev_tb"))
            and bool(self.raw_event.get("ev_tb_catalog"))
            and bool(self.raw_event.get("ev_tb_schema"))
            and bool(self.raw_event.get("ev_tb_table"))
            and bool(self.raw_event.get("st_code"))
            and bool(self.raw_event.get("st_done"))
            and bool(self.raw_event.get("st_dur_ms"))
        )

    @public
    def to_default_asset_events(
        self,
        dagster_sdf_translator: DagsterSdfTranslator = DagsterSdfTranslator(),
        context: Optional[OpExecutionContext] = None,
    ) -> Iterator[SdfDagsterEventType]:
        """Convert an sdf CLI event to a set of corresponding Dagster events.

        Args:
            context (Optional[OpExecutionContext]): The execution context.

        Returns:
            Iterator[Union[Output, AssetMaterialization]]:
                A set of corresponding Dagster events.

                In a Dagster asset definition, the following are yielded:
                - Output for refables (e.g. models)

                In a Dagster op definition, the following are yielded:
                - AssetMaterialization for refables (e.g. models)
        """
        if not self.is_result_event:
            return

        is_success = self.raw_event["st_code"] == "succeeded"
        if not is_success:
            return

        selected_output_names: AbstractSet[str] = (
            context.selected_output_names if context else set()
        )

        catalog = self.raw_event["ev_tb_catalog"]
        schema = self.raw_event["ev_tb_schema"]
        table = self.raw_event["ev_tb_table"]
        purpose = self.raw_event["ev_tb_purpose"]
        dialect = self.raw_event["ev_tb_dialect"]

        # If assets are selected, only yield events for selected assets
        if len(selected_output_names) > 0:
            exists = exists_in_selected(
                catalog,
                schema,
                table,
                purpose,
                dialect,
                selected_output_names,
                dagster_sdf_translator.settings.enable_asset_checks,
            )
            if not exists:
                return

        table_id = self.raw_event["ev_tb"]
        default_metadata = {
            DAGSTER_SDF_TABLE_ID: table_id,
            DAGSTER_SDF_CATALOG_NAME: catalog,
            DAGSTER_SDF_SCHEMA_NAME: schema,
            DAGSTER_SDF_TABLE_NAME: table,
            DAGSTER_SDF_PURPOSE: purpose,
            DAGSTER_SDF_DIALECT: dialect,
            "Execution Duration": self.raw_event["st_dur_ms"] / 1000,
            "Materialized From Cache": True if self.raw_event["st_done"] == "cached" else False,
        }
        asset_key = dagster_sdf_translator.get_asset_key(catalog, schema, table)
        if purpose == "model":
            has_asset_def = bool(context and context.has_assets_def)
            event = (
                Output(
                    value=None,
                    output_name=dagster_name_fn(
                        catalog,
                        schema,
                        table,
                    ),
                    metadata=default_metadata,
                )
                if has_asset_def
                else AssetMaterialization(
                    asset_key=asset_key,
                    metadata=default_metadata,
                )
            )

            yield event
        elif purpose == "test" and dagster_sdf_translator.settings.enable_asset_checks:
            passed = self.raw_event["st_verdict"] == "passed"
            asset_check_key = dagster_sdf_translator.get_check_key_for_test(catalog, schema, table)
            # if the asset key is the same as the asset check key, then the test is not a table / column test
            if asset_key == asset_check_key.asset_key:
                return
            metadata = {
                **default_metadata,
                "status_verdict": self.raw_event["st_verdict"],
            }
            yield AssetCheckResult(
                passed=passed,
                asset_key=asset_check_key.asset_key,
                check_name=asset_check_key.name,
                metadata=metadata,
                severity=AssetCheckSeverity.ERROR,
            )
        else:
            return



import os
from pathlib import Path
from typing import (
    AbstractSet,
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import dagster._check as check
import polars as pl
from dagster import (
    AssetCheckSpec,
    AssetKey,
    AssetObservation,
    AssetSpec,
    OpExecutionContext,
    TableColumn,
    get_dagster_logger,
)
from dagster._core.definitions.metadata import (
    CodeReferencesMetadataSet,
    CodeReferencesMetadataValue,
    LocalFileCodeReference,
    TableColumnConstraints,
    TableMetadataSet,
    TableSchema,
)
from dagster._core.definitions.tags import build_kind_tag
from dagster._record import IHaveNew, record_custom

from dagster_sdf.asset_utils import exists_in_selected, get_info_schema_dir, get_output_dir
from dagster_sdf.constants import (
    DAGSTER_SDF_CATALOG_NAME,
    DAGSTER_SDF_DIALECT,
    DAGSTER_SDF_PURPOSE,
    DAGSTER_SDF_SCHEMA_NAME,
    DAGSTER_SDF_TABLE_ID,
    DAGSTER_SDF_TABLE_NAME,
    DEFAULT_SDF_WORKSPACE_ENVIRONMENT,
    SDF_INFORMATION_SCHEMA_TABLES_STAGE_COMPILE,
    SDF_INFORMATION_SCHEMA_TABLES_STAGE_PARSE,
)
from dagster_sdf.dagster_sdf_translator import DagsterSdfTranslator
from dagster_sdf.sdf_event_iterator import SdfDagsterEventType

logger = get_dagster_logger()


@record_custom(checked=False)
class SdfInformationSchema(IHaveNew):
    """A class to represent the SDF information schema.

    The information schema is a set of tables that are generated by the sdf cli on compilation.
    It can be queried directly via the sdf cli, or by reading the parquet files that live in the
    `sdftarget` directory.

    This class specifically interfaces with the tables and columns tables, which contain metadata
    on their upstream and downstream dependencies, as well as their schemas, descriptions, classifiers,
    and other metadata.

    Read more about the information schema here: https://docs.sdf.com/reference/sdf-information-schema#sdf-information-schema

    Args:
        workspace_dir (Union[Path, str]): The path to the workspace directory.
        target_dir (Union[Path, str]): The path to the target directory.
        environment (str, optional): The environment to use. Defaults to "dbg".
    """

    workspace_dir: Path
    target_dir: Path
    environment: str
    information_schema_dir: Path
    information_schema: Dict[str, pl.DataFrame]

    def __new__(
        cls,
        workspace_dir: Union[Path, str],
        target_dir: Union[Path, str],
        environment: str = DEFAULT_SDF_WORKSPACE_ENVIRONMENT,
    ):
        check.inst_param(workspace_dir, "workspace_dir", (str, Path))
        check.inst_param(target_dir, "target_dir", (str, Path))
        check.str_param(environment, "environment")

        workspace_dir = Path(workspace_dir)
        target_dir = Path(target_dir)

        information_schema_dir = get_info_schema_dir(target_dir, environment)
        check.invariant(
            information_schema_dir.exists(),
            f"Information schema directory {information_schema_dir} does not exist.",
        )

        return super().__new__(
            cls,
            workspace_dir=workspace_dir,
            target_dir=target_dir,
            environment=environment,
            information_schema_dir=information_schema_dir,
            information_schema={},
        )

    def read_table(
        self,
        table_name: Literal["tables", "columns", "table_lineage", "column_lineage", "table_deps"],
    ) -> pl.DataFrame:
        check.invariant(
            table_name
            in SDF_INFORMATION_SCHEMA_TABLES_STAGE_COMPILE
            + SDF_INFORMATION_SCHEMA_TABLES_STAGE_PARSE,
            f"Table `{table_name}` is not valid information schema table."
            f" Select from one of {SDF_INFORMATION_SCHEMA_TABLES_STAGE_COMPILE + SDF_INFORMATION_SCHEMA_TABLES_STAGE_PARSE}.",
        )

        return self.information_schema.setdefault(
            table_name, pl.read_parquet(self.information_schema_dir.joinpath(table_name))
        )

    def build_sdf_multi_asset_args(
        self, dagster_sdf_translator: DagsterSdfTranslator
    ) -> Tuple[
        Sequence[AssetSpec],
        Sequence[AssetCheckSpec],
    ]:
        table_id_to_dep: Dict[str, AssetKey] = {}
        table_id_to_upstream: Dict[str, Set[AssetKey]] = {}
        asset_specs: Sequence[AssetSpec] = []
        asset_checks: Sequence[AssetCheckSpec] = []
        origin_remote_tables: Set[str] = set()

        # Step 0: Filter out system and external-system tables
        table_deps = self.read_table("table_deps").filter(
            ~pl.col("purpose").is_in(["system", "external-system"])
        )
        table_columns: Dict[str, List[TableColumn]] = {}
        try:
            table_columns = self.get_columns()
        except Exception:
            logger.warn("Column information schema table could not be read.")

        # Step 1: Build Map of Table Deps to Rows
        table_rows_deps = {row["table_id"]: row for row in table_deps.rows(named=True)}
        # Step 2: Build Dagster Asset Deps
        for table_row in table_deps.rows(named=True):
            # Iterate over the meta column to find the dagster-asset-key
            if len(table_row["meta"]) > 0:
                for meta_map in table_row["meta"]:
                    # If the meta_map has a key of dagster-asset-key, add it to the deps
                    if meta_map["keys"] == "dagster-asset-key":
                        dep_asset_key = meta_map["values"]
                        table_id_to_dep[table_row["table_id"]] = AssetKey(dep_asset_key)
                    elif meta_map["keys"] == "dagster-depends-on-asset-key":
                        dep_asset_key = meta_map["values"]
                        # Currently, we only support one upstream asset
                        table_id_to_upstream.setdefault(table_row["table_id"], set()).add(
                            AssetKey(dep_asset_key)
                        )
                    elif table_row["origin"] == "remote":
                        origin_remote_tables.add(table_row["table_id"])
            elif table_row["origin"] == "remote":
                origin_remote_tables.add(table_row["table_id"])

        # Step 3: Build Dagster Asset Outs and Internal Asset Deps
        for table_row in table_deps.rows(named=True):
            asset_key = dagster_sdf_translator.get_asset_key(
                table_row["catalog_name"], table_row["schema_name"], table_row["table_name"]
            )
            code_references = None
            if dagster_sdf_translator.settings.enable_code_references:
                code_references = self._extract_code_ref(table_row)
            metadata = {
                **TableMetadataSet(
                    column_schema=TableSchema(
                        columns=table_columns.get(table_row["table_id"], []),
                    ),
                    relation_identifier=table_row["table_id"],
                ),
                **(code_references if code_references else {}),
                DAGSTER_SDF_TABLE_ID: table_row["table_id"],
                DAGSTER_SDF_CATALOG_NAME: table_row["catalog_name"],
                DAGSTER_SDF_SCHEMA_NAME: table_row["schema_name"],
                DAGSTER_SDF_TABLE_NAME: table_row["table_name"],
                DAGSTER_SDF_PURPOSE: table_row["purpose"],
                DAGSTER_SDF_DIALECT: table_row["dialect"],
            }
            # If the table is a annotated as a dependency, we don't need to create an output for it
            if (
                table_row["table_id"] not in table_id_to_dep
                and table_row["table_id"] not in origin_remote_tables
            ):
                dependencies = {
                    table_id_to_dep[dep]
                    if dep
                    in table_id_to_dep  # If the dep is a dagster dependency, use the meta asset key
                    else dagster_sdf_translator.get_asset_key(
                        table_rows_deps[dep]["catalog_name"],
                        table_rows_deps[dep]["schema_name"],
                        table_rows_deps[dep]["table_name"],
                    )  # Otherwise, use the translator to get the asset key
                    for dep in table_row["depends_on"]
                    if dep not in origin_remote_tables and dep in table_deps["table_id"]
                }.union(table_id_to_upstream.get(table_row["table_id"], set()))
                asset_specs.append(
                    AssetSpec(
                        key=asset_key,
                        deps=dependencies,
                        description=dagster_sdf_translator.get_description(
                            table_row,
                            self.workspace_dir,
                            get_output_dir(self.target_dir, self.environment),
                        ),
                        metadata=metadata,
                        tags={
                            **build_kind_tag("sdf"),
                        },
                        skippable=True,
                    )
                )
                # This registers an asset check on all inner tables, since SDF will execute all tests as a single query (greedy approach)
                # If no table or column tests are registered, they will simply be skipped
                if dagster_sdf_translator.settings.enable_asset_checks:
                    test_name = "test"
                    asset_checks.append(
                        AssetCheckSpec(
                            name=test_name,
                            asset=asset_key,
                        )
                    )
        return asset_specs, asset_checks

    def get_columns(self) -> Dict[str, List[TableColumn]]:
        columns = self.read_table("columns")[
            ["table_id", "column_id", "classifiers", "column_name", "datatype", "description"]
        ]
        table_columns: Dict[str, List[TableColumn]] = {}
        for row in columns.rows(named=True):
            if row["table_id"] not in table_columns:
                table_columns[row["table_id"]] = []
            table_columns[row["table_id"]].append(
                TableColumn(
                    name=row["column_name"],
                    type=row["datatype"] if row["datatype"] else "unknown",
                    description=row["description"],
                    constraints=TableColumnConstraints(
                        other=row["classifiers"] if row["classifiers"] else []
                    ),
                )
            )
        return table_columns

    def _extract_code_ref(
        self, table_row: Dict[str, Any]
    ) -> Union[CodeReferencesMetadataSet, None]:
        code_references = None
        # Check if any of the source locations are .sql files, return the first one
        loc = (
            next(
                (
                    source_location
                    for source_location in table_row["source_locations"]
                    if source_location.endswith(".sql")
                ),
                None,
            )
            or next(
                (
                    source_location
                    for source_location in table_row["source_locations"]
                    if source_location.endswith(".sdf.yml")
                ),
                None,
            )
            or "workspace.sdf.yml"
        )
        code_references = CodeReferencesMetadataSet(
            code_references=CodeReferencesMetadataValue(
                code_references=[
                    LocalFileCodeReference(file_path=os.fspath(self.workspace_dir.joinpath(loc)))
                ]
            )
        )
        return code_references

    def stream_asset_observations(
        self,
        dagster_sdf_translator: DagsterSdfTranslator,
        context: Optional[OpExecutionContext] = None,
    ) -> Iterator[SdfDagsterEventType]:
        selected_output_names: AbstractSet[str] = (
            context.selected_output_names if context else set()
        )
        tables = self.read_table("tables")
        tables = tables.filter(~pl.col("purpose").is_in(["system", "external-system"]))
        table_columns = self.get_columns()
        for table_row in tables.rows(named=True):
            # If selected outputs are expected, only yield events for selected outputs
            if len(selected_output_names) > 0:
                exists = exists_in_selected(
                    table_row["catalog_name"],
                    table_row["schema_name"],
                    table_row["table_name"],
                    table_row["purpose"],
                    table_row["dialect"],
                    selected_output_names,
                    dagster_sdf_translator.settings.enable_asset_checks,
                )
                if not exists:
                    continue

            asset_key = dagster_sdf_translator.get_asset_key(
                table_row["catalog_name"], table_row["schema_name"], table_row["table_name"]
            )
            code_references = None
            if dagster_sdf_translator.settings.enable_code_references:
                code_references = self._extract_code_ref(table_row)
            metadata = {
                **TableMetadataSet(
                    column_schema=TableSchema(
                        columns=table_columns.get(table_row["table_id"], []),
                    ),
                    relation_identifier=table_row["table_id"],
                ),
                **(code_references if code_references else {}),
                DAGSTER_SDF_TABLE_ID: table_row["table_id"],
                DAGSTER_SDF_CATALOG_NAME: table_row["catalog_name"],
                DAGSTER_SDF_SCHEMA_NAME: table_row["schema_name"],
                DAGSTER_SDF_TABLE_NAME: table_row["table_name"],
                DAGSTER_SDF_PURPOSE: table_row["purpose"],
                DAGSTER_SDF_DIALECT: table_row["dialect"],
            }
            yield AssetObservation(
                asset_key=asset_key,
                description=dagster_sdf_translator.get_description(
                    table_row, self.workspace_dir, get_output_dir(self.target_dir, self.environment)
                ),
                metadata=metadata,
            )

    def is_compiled(self) -> bool:
        for table in SDF_INFORMATION_SCHEMA_TABLES_STAGE_COMPILE:
            if not any(self.information_schema_dir.joinpath(table).iterdir()):
                return False
        return True

    def is_parsed(self) -> bool:
        for table in SDF_INFORMATION_SCHEMA_TABLES_STAGE_PARSE:
            if not any(self.information_schema_dir.joinpath(table).iterdir()):
                return False
        return True
